@@ -11,7 +11,7 @@ setup(
 #    Extension("test", ["test.pyx"]),
     Extension("func_pointers", ["func_pointers.pyx"]),
 #    Extension("inplace", ["inplace.pyx"]),
-    Extension("withGIL", ["withGIL.pyx"]),
+#    Extension("withGIL", ["withGIL.pyx"]),
     Extension("class_members", ["class_members.pyx"]),
 #    Extension("inherit_bug", ["inherit_bug.pyx"]),
     Extension("override", ["override.pyx"]),
@@ -19,6 +19,9 @@ setup(
     Extension("submodule.test",       ["submodule/test.pyx"]),
     Extension("errors",       ["errors.pyx"]),
     Extension("cpdef",       ["cpdef.pyx"]),
+    Extension("range",       ["range.pyx"]),
+    Extension("ints",       ["ints.pyx"]),
+    Extension("detect_override",       ["detect_override.pyx"]),
     ],
   cmdclass = {'build_ext': build_ext},
 #  include_dirs = "/System/Library/Frameworks/Python.framework/Versions/2.3/include/python2.3/"
